# olojs.Store module.
- **Version:** 0.2
- **Author:** Marcello Del Buono <m.delbuono@onlabs.org>
- **License:** MIT
- **Content:**
    - [class Store][Store]
    - [class Store.Document][Document]
    - [class Store.Document.Item][Item]
    - [class Store.Document.Dict][Dict]
    - [class Store.Document.List][List]
    - [class Store.Document.Text][Text]
    - [class Store.Document.Change][Change]
    - [class Store.Document.Subscription][Subscription]
  
## Store abstract class
A `Store` object is a frontend interface to a database containing JSON documents.
This class is an interface that can be implemented for several backends.
  
### Store.prototype.connect(credentials) - async method
This method establishes a connection with the backend.

###### Parameters
- `credentials` : is an implementation-specific object that identified the user for access control.
  
###### Resolves
- `undefined` if the connection was successfull
  
###### Rejects
- `Error` if the connection attempt failed
  
> This method calls internally the implementation-specific
> method `.__connect` (see below).
  
### Document.prototype.__connect(credentials) - async virtual method
The `__connect` method implementations should
- establish a connection to the backend
- log-in the user to the backend
  
###### Parameters
- `credentials` : implementation-specific object used to log-in
  
###### Resolves
- the logged-in user id if the connection was successfull
  
###### Rejects
- `Error` if either connection or login fails
  
### User.prototype.getUserRole(docId) - async method
The `.getUserRole` method retrieves the access control role of the 
logged-in user on a specific document.
###### Parameters
- `docId` the id of the document for which we want to know the user role
  
###### Resolves
- `roles.OWNER` if the user has write permissions on the document data and metadata
- `roles.WRITER` if the user has write permissions on the document data
- `roles.READER` if the user has read-only permissions on the document
- `roles.NONE` if the user has no permissions on the document
  
> * The root-level keys of a document are meta-data, while the root key `data`
> contains the document data.
> * By default this method resolves always `roles.OWNER` and it shoul
> be overridden by the implementations to define a proper access control.
  
### Store.prototype.connected - getter
###### Returns
- `true` if the store-backend connection is active
- `false` otherwise
  
> This getter calls internally the implementation-specific
> method `.__connected` (see below).
  
### Document.prototype.__connected() - async virtual method
###### Returns
- `true` if the store-backend connection is active
- `false` otherwise
  
### Store.prototype.getDocument(docId) - async method
###### Parameters
- `docId` : a remote document id
  
###### Resolves
- the [Document][] object with the given id
- the same object when passing the same id (documents are cached)
  
###### Rejects
- `ReadPermissionError` if the user has no read permissions on the
  requested document
  
### Store.prototype.disconnect() - async method
Closes the connection to the backend.  
Closes all the open documents.  
Cleares the documents cache.  
  
###### Resolves
- `undefined` when the connection is established
  
###### Rejects
- `Error` if the disconnection failed
  
> This method calls internally the implementation-specific
> method `.__disconnect` (see below).
  
### Document.prototype.__disconnect() - async virtual method
Closes the connection to the backend.

###### Resolves
- `undefined` when the connection is established
  
###### Rejects
- `Error` if the disconnection failed
  
## Document abstract class
A `Document` object represents a JSON document contained in a [Store][].
  
### new Document(store, docId) - constructor
###### Parameters
- `store` : is the [Store][] containing this document
- `docId` : is the id of this document
  
> You shouldn't call this constructor directly.
> Use `Store.prototype.getDocument` instead.
  
### Store.Document.prototype.__init() - async virtual method
This method is called by the `Store.prototype.getDocument` method 
and should contain the asynchronous initialization
of the `Store.Document` implmentations.
###### Should return
- `undefined` always
  
### Document.prototype.store - getter
###### Returns
- the [Store][] that contains this document
  
### Document.prototype.id - getter
###### Returns
- the id of this document
  
### Document.prototype.open() - async method
Fetches the remote document and subscribes to change notifications.
If the remote document doesn't exists, creates it.
  
###### Resolves
- `undefined` when the document has been fetched and subscribed
  
###### Rejects
- `errors.ReadPermissionError` if the user has no read permissions on the document
- `errors.WritePermissionError` if the document doesn't exist and 
  the user has no write permissions on the document
  
> This method calls internally the implementation-specific
> methoda `.__fetch` and `.__create` (see below).
  
### Store.Document.prototype.__fetch() - async virtual method
Fetches the remote document and subscribes to change notifications.
  
###### Should Resolve
- the document content as JSON object if the operation was successfull
- `null` if the document doesn't exist
  
###### Should Reject
- `Error` if the operation failed
  
### Store.Document.prototype.__create(content) - async virtual method
Creates a new empty document and initializes its content.
###### Parameters
- `content` the initial content of the document
  
###### Should Resolve
- `undefined` if the operation was successfull
  
###### Should Reject
- `Error` if the operation failed
  
### Document.prototype.userRole - getter
This getter holds the result of `Store.prototype.getUserRole`
after the document is open.

###### Returns
- `roles.OWNER` if the user has write permissions on this document data and metadata
- `roles.WRITER` if the user has write permissions on this document data
- `roles.READER` if the user has read-only permissions on this document
- `roles.OWNER` if the user has no permissions on this document
  
### Document.prototype.get(...path) - method
Creates a pointer to the to a document item.
The returned pointer object is mutable:
- It behaves like a [Dict][] object when the target is a mapping object
- It behaves like a [List][] object when the target is an array
- It behaves like a [Text][] object when the target is a string
- It behaves like an [Item][] object when the target is a primitive value
If the target type changes, the pointer object will adjust its behaviour.
This is achieved by using [Proxy][] objects.
  
###### Parameters
- `path` : a document item path
  
###### Returns
- the pointer to the document item
- the same pointer when passing the same path (caches items)
  
###### Throws
- `errors.DocumentClosedError` if the document is not open
- `errors.ReadPermissionError` if the user has no read permissions on the document
> The `.get` method caches the returned value, therefore it will always
> return the same pointer when passing the same path.
  
### Document.prototype.close() - async method
Closes the connection with the document, cancels all the
subscriptions and clears the items cache.

###### Resolves
- `undefined` if the document has been successfully closed
  
###### Rejects
- `Error` if the closing didn't succeed.
  
> This method calls internally the implementation-specific
> method `.__close` (see below).
  
### Store.Document.prototype.__close() - async virtual method
The `.__close` method implementations should disconnect from the
backend document and cancel the change subscription.
###### Should resolve
- `undefined` if the document was successfully closed
  
###### Should reject
- `Error` if the document could not be closed
  
### Store.Document.prototype.__getItemValue(path) - virtual method
The `.__getItemValue` method implementations should retrieve the value of a
document item.
###### Parameters
- `path` : the document item path
  
###### Should return
- the item raw value if it exists
- `null` if the item path doesn't exist
  
> The implementation can trust that
> - this document is open
> - `path` is always a [Path][] instance,
> - the user has read permissions on the document path
  
### Store.Document.prototype.setDictItem(dictPath, key, newValue) - virtual method
The `.__setDictItem` method implementations should change the value of
of a dictionary item.
As a consequence of this change, `this._dispatch` should be called.
###### Parameters
- `dictPath` : the path to the dictionary containing the target item
- `key` : the key of the target item
- `newValue` : the value to assign to the target item
  
###### Should return
- `undefined` if the assignment was successful
  
> The implementation can trust that:
> - this document is open
> - `dictPath` is always a [Path][] instance
> - `dictPath` points always to a dictionary item
> - `newValue` is a valid JSON value
> - the user has write permissions on the document path
  
### Store.Document.prototype.__removeDictItem(dictPath, key) - virtual method
The `.__removeDictItem` method implementations should remove a dictionary item.
As a consequence of this change, `this._dispatch` should be called.
###### Parameters
- `dictPath` : the path to the dictionary containing the target item
- `key` : the key of the target item
  
###### Should return
- `undefined` if the item was successfully removed
  
> The implementation can trust that:
> - this document is open
> - `path` is always a [Path][] instance
> - `path` points always to a dictionary item
> - the user has write permissions on the document path
  
### Store.Document.prototype.__setListtem(listPath, index, newItem) - virtual method
The `.__setListItem` method implementations should change the value of a list item.
As a consequence of this change, `this._dispatch` should be called.
###### Parameters
- `listPath` : the path to the list containing the target item
- `index` : the position of the target item
- `newItem` : the value to assign to the target item
  
###### Should return
- `undefined` if the assignment was successfull
  
> The implementation can trust that:
> - this document is open
> - `path` is always a [Path][] instance
> - `path` points always to an array item
> - `newItem` is a valid JSON value
> - the user has write permissions on the document path
  
### Store.Document.prototype.__insertListItem(listPath, index, newItem) - virtual method
The `.__insertListItem` method implementations should insert an item
in a list and shift the other items.
As a consequence of this change, `this._dispatch` should be called.
###### Parameters
- `listPath` : the path to the list containing the target item
- `index` : the position where the item should be placed
- `newItem` : the item to be inserted
  
###### Should return
- `undefined` if the item was successfully inserted
  
> The implementation can trust that:
> - this document is open
> - `path` is always a [Path][] instance
> - `path` points always to an array item
> - `index` is a valid array index
> - `newItem` is a valid JSON value
> - the user has write permissions on the document path
  
### Store.Document.prototype.__removeListItem(listPath, index) - virtual method
The `.__removeListItem` method implementations should remove an item
from its parent list and shift the other items.
As a consequence of this change, `this._dispatch` should be called.
###### Parameters
- `listPath` : the path to the list containing the target item
- `index` : the position of the target item
  
###### Should return
- `undefined` if the item was successfully removed
  
> The implementation can trust that:
> - this document is open
> - `path` is always a [Path][] instance
> - `path` points always to an array item
> - `index` is a valid array index
> - the user has write permissions on the document path
  
### Store.Document.prototype.__insertTextString(textPath, index, subString) - virtual method
The `.__insertTextString` method implementations should insert a ssub-string
in a text item.
As a consequence of this change, `this._dispatch` should be called.
###### Parameters
- `textPath` : the path to the target text item
- `index` : the position where the sub-string should be inserted
- `subString` : the string to be inserted
  
###### Should return
- `undefined` if the sub-string was successfully inserted
  
> The implementation can trust that:
> - this document is open
> - `path` is always a [Path][] instance
> - `path` points always to a text item
> - `index` is a valid string index
> - the user has write permissions on the document path
  
### Store.Document.prototype.__removeTextString(textPath, index, count) - virtual method
The `.__removeTextString` method implementations should remove a sub-string
from a text item,
As a consequence of this change, `this._dispatch` should be called.
###### Parameters
- `textPath` : the path to the target text item
- `index` : the position where to start removing characters
- `count` : the number of characters to be removed
  
###### Should return
- `undefined` if the sub-string was successfully removed
  
> The implementation can trust that:
> - this document is open
> - `path` is always a [Path][] instance
> - `path` points always to a text item
> - `index` is a valid string index
> - `count` is a valid number of characters
> - the user has write permissions on the document path
  
## Item class
An `Item` object represents a value contained in a JSON document.
  
### new Item(doc, path) - constructor
###### Parameters
- `doc` : is the [Document][] that contains the item
- `path` is the [Path][] object or path literal or path array that
  defines the item position inside the document
  
> This constructor should not be called directly.
> Use `Document.prototype.get` insetead.
  
### Item.prototype.doc - getter
###### Returns
- the [Document][] object that contains this item.
  
### Item.prototype.path - getter
###### Returns
- the [Path][] object that defines the item position inside its document.
  
### Item.prototype.type - getter
###### Returns
- `"dict"` if the underlying data is an object
- `"list"` if the underlying data is an array
- `"text"` if the underlying data is a string
- `"numb"` if the underlying data is a number
- `"bool"` if the underlying data is a boolean
- `"none"` if the item doesn't exist
- `"none"` if the item.readable is false
  
### Item.prototype.readonly - getter
###### Returns
- `true` if the user has no write permissions on this item
- `false` otherwise
  
### Item.prototype.get(subPath) - method
Retrieves a sub-item of this item.
In other words it is a shortcut for `item.doc.get([item.path, subPath])`
###### Parameters
- `subPath` : the relarive path of the target sub-item
  
###### Returns
- a pointer to the sub-item.
  
### Item.prototype.value - getter
Retrieves this item value.
###### Returns
- a deep copy of the underlying readable data.
- `null` if this item path doesn't exist
- `null` if this.readable is false
  
> Changing the returned value will not change the document data.
  
### Item.prototype.value - setter
Changes the underlying data to a deep copy of the passed value.
###### Returns
- `undefined` if the assignment was successfull
  
###### Throws
- `errors.DocumentClosedError` if the containing document is not open
- `errors.ReadPermissionError` if the user has no read permissions
- `errors.WritePermissionError` if the user has no write permissions
- `Error` if trying to set the document root item (`doc.get()`)
  to something other than a dictionary
  
### Item.prototype.subscribe(callback) - method
Registers a callback that will be invoked with a [Change][] object
as parameter, each time the item value changes.
###### Returns
a [Subscription][] object
###### Throws
- `errors.DocumentClosedError` if the document is not open
  
## Dict class
A `Dict` is an [Item][] object that represents a dictionary contained in
a document.
  
### Dict.prototype.keys - getter
###### Returns
- an array with all the keys contained in the dictionary.
  
###### Throws
- `errors.DocumentClosedError` if the document is not open
- `errors.ReadPermissionError` if the user has no read pemissions
  
### Dict.prototype.set(key, value)
Assigns a value to a key.
###### Parameters
- `key` : the key to be assigned a new value
- `value` : the new value
  
###### Returns
- `udefined` if the assigment was successfull
  
###### Throws
- `errors.DocumentClosedError` if the document is not open
- `errors.WritePermissionError` if the user has no write pemissions
- `Error` if the new value is not a valid JSON object or if it is null
  
### Dict.prototype.remove(key)
Removes a key
###### Parameters
- `key` : the key to be removed
  
###### Returns
- `udefined` if the key was successfully removed
  
###### Throws
- `errors.DocumentClosedError` if the document is not open
- `errors.WritePermissionError` if the user has no write pemissions
  
## List class
A `List` is an [Item][] object that represents an array object contained
in a JSON document.
  
### List.prototype.size - getter
###### Returns
- the number of items in the list.
  
###### Throws
- `errors.DocumentClosedError` if the document is not open
- `errors.ReadPermissionError` if the user has no read pemissions
  
### List.prototype.set(index, item) - method
Changes an item value.
###### Parameters
- `index` : the position of the item to be assigned a value
- `item` : the new item value
  
###### Returns
- `undefined` if the assignment was successfull
  
###### Throws
- `errors.DocumentClosedError` if the document is not open
- `errors.WritePermissionError` if the user has no write pemissions
- `Error` if the new item value is not a valid JSON object or if it is null
- `Error` if index is not a valid number.
  
> If index is negative, it will be considered as relative to the end
> of the list.
  
### List.prototype.insert(index, ...items) - method
Inserts sub-items to a list.
###### Parameters
- `index` : the position of the item to be assigned a value
- `...items` : the sub-items to be added
  
###### Returns
- `undefined` if the items were successfully inserted
  
###### Throws
- `errors.DocumentClosedError` if the document is not open
- `errors.WritePermissionError` if the user has no write pemissions
- `Error` if the new item value is not a valid JSON object or if it is null
- `Error` if index is not a valid number.
  
> If index is negative, it will be considered as relative to the end
> of the list.
  
### List.prototype.append(...items)
Shortcut for `list.insert(list.size, ...items)`
  
### List.prototype.remove(index, count)
Remove items from a list.
###### Parameters
- `index` : the position of the first item to be removed
- `count` : the number of items to be removed (default to 1)
  
###### Returns
- `undefined` if the items were successfully removed
  
###### Throws
- `errors.DocumentClosedError` if the document is not open
- `errors.WritePermissionError` if the user has no write pemissions
- `Error` if index is not a valid number.
  
> If index is negative, it will be considered as relative to the end
> of the list.
  
## Text class
A `Text` is an [Item][] object that represents a string contained in
a JSON document.
  
### Text.prototype.size - getter
###### Returns
- the length of the string.
  
###### Throws
- `errors.DocumentClosedError` if the document is not open
- `errors.ReadPermissionError` if the user has no read pemissions
  
### Text.prototype.insert(index, subString) - method
Inserts a sub-string in a text item.
###### Parameters
- `index` : the position where the sub-string should be inserted
- `subString` : the string to be inserted
  
###### Returns
- `undefined` if the sub-string was successfully inserted
  
###### Throws
- `errors.DocumentClosedError` if the document is not open
- `errors.WritePermissionError` if the user has no write pemissions
- `Error` if index is not a valid number.
  
> If index is negative, it will be considered as relative to the end
> of the text.
  
### Text.prototype.append(subString)
Shortcut for `text.insert(text.size, subString)`
  
### Text.prototype.remove(index, count)
Removes a sub-string from a text item.
###### Parameters
- `index` : the position of the first character to be removed
- `count` : the number of characters to be removed (defaults to 1)
  
###### Returns
- `undefined` if the sub-string was successfully removed
  
###### Throws
- `errors.DocumentClosedError` if the document is not open
- `errors.WritePermissionError` if the user has no write pemissions
- `Error` if index is not a valid number.
  
> If index is negative, it will be considered as relative to the end
> of the text.
  
## Change class
A `Change` object represents a change occurred in a [Document][] [Item][].
It is passed to the change listeners attached to an [Item][] object
via the `Item.prototype.subscribe` method.
###### Properties
- `change.path`: the [Path][] object defining the position of the item that changed
- `change.removed`: the removed (old value) of the item
- `change.inserted`: the added (new value) of the item
  
> - A change object with the `removed` and `inserted` property both non null, represents a `set` change.
> - A change object with null `inserted` property, represents a `remove` change.
> - A change object with null `removed` property, represents an `insert` change.
  
### Change.prototype.SubChange(subPath)
Genertates the Change object relative to a sub-item.
###### Parameters
- `path` : the subPath of the target item
  
###### Returns
- a [Change][] object representing the effects of this change
  on the sub-item.
- `null` if the sub-item is not affected by this change
  
## Subscription class
Represent a change subscription and it is returned by the `Item.prototype.subscribe` method.
###### Properties
- `subscription.doc`: is the [Document][] containing the observed [Item][]
- `subscription.path`: is the [Path][] of the observed [Item][]
- `subscription.callback`: is the function that gets invoked when the item changes
  
### Subscription.prototype.cancel() - method
Cancels the subscription. As a consequence, furute changes to the observed item
will no longer be notified to `subscription.callback`.
###### Returns
- `undefined` if the subscription was successfully cancelled
  
[Store]: #store-abstract-class
[Document]: #document-abstract-class
[User]: #user-class
[Item]: #item-class
[Dict]: #dict-class
[List]: #list-class
[Text]: #text-class
[Change]: #change-class
[Subscription]: #subscription-class
[AbstractBackend]: ./AbstractBackend.md
[Path]: ./Path.md#path-class
[Proxy]: https://developer.mozilla.org/it/docs/Web/JavaScript/Reference/Global_Objects/Proxy
  
